"""
Create partial model definitions, sort of like Partial<> in TypeScript

Sources:
https://github.com/pydantic/pydantic/issues/3120#issuecomment-1528030416
https://stackoverflow.com/a/76560886
"""
from copy import deepcopy
from typing import Any, Dict, Generic, Optional, TypeVar
from typing import Type

from pydantic import Extra, create_model, BaseModel as PydanticBaseModel, Field
from pydantic.fields import ModelField

# pylint: disable=missing-function-docstring


T = TypeVar("T", bound="BaseModel")


class BaseModel(PydanticBaseModel, Generic[T]):
    @classmethod
    def partial(cls, **kwargs):
        partial_model = make_partial_model(cls)
        return partial_model(**{k: v for (k, v) in kwargs.items() if v is not None})

    @classmethod
    def from_model(cls, model: Optional[T] = None) -> Optional[T]:
        if not model:
            return None
        return cls.from_orm(model)

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]] = None) -> Optional[T]:
        if not obj:
            return None
        return cls.parse_obj(obj)

    @classmethod
    def from_json(cls, json: Optional[str] = None) -> Optional[T]:
        if not json:
            return None
        return cls.parse_raw(json)

    @classmethod
    def model_fields(cls) -> Dict[str, Any]:
        return cls.__fields__

    def model_dump(self, exclude_none: bool = False) -> Dict[str, Any]:
        return self.dict(exclude_none=exclude_none)

    def model_dump_json(self, exclude_none: bool = False, exclude_unset: bool = False) -> str:
        return self.json(exclude_none=exclude_none, exclude_unset=exclude_unset)

    class Config:
        orm_mode = True
        allow_population_by_field_name = True
        allow_mutation = True


class RootModel(BaseModel):
    def __getattr__(self, name):
        return self.__root__.__getattribute__(name)


def make_field_optional(field: ModelField, default: Any = None) -> ModelField:
    new = deepcopy(field)
    new.default = default
    new.required = False
    new.annotation = Optional[field.annotation]  # type: ignore
    return new


BaseModelT = TypeVar("BaseModelT", bound=BaseModel)


class OverridePartialModelMethods(BaseModel):
    """
    Base class of partial models
    """

    def dict(self, *args, **kwargs):
        """
        same as BaseModel.dict(), but only for fields present in the model.
        """
        return dict(super().dict(*args, exclude_unset=True, **kwargs).items())

    def json(self, *args, **kwargs):
        """
        same as BaseModel.json(), but only for fields present in the model.
        """
        return super().json(*args, exclude_unset=True, exclude_none=True, **kwargs)

    class Config:
        extra = Extra.forbid


def make_partial_model(model: Type[BaseModelT]) -> Type[BaseModelT]:
    __fields__ = {
        field_name: make_field_optional(field_info)
        for field_name, field_info in model.__fields__.items()
    }
    m: Type[BaseModelT] = create_model(
        f"Partial{model.__name__}",
        __base__=(model, OverridePartialModelMethods),  # type: ignore
        __module__=model.__module__,
    )
    m.__fields__ = __fields__
    return m


class Entity(BaseModel):
    """
    Base class for all entities
    """

class hiddenstr(str):
    pass


# !!!!! DO NOT EVER CHANGE THIS FILE MANUALLY -- AUTOGENERATED by orm2pydantic.py !!!!!
# Generated from module egp_api_backend.server.internal.orm_models in the egp-api-backend package
# To regenerate this file, run:
# scaleapi/packages/egp-api-backend/scripts/orm2pydantic.sh

import datetime
import enum
from typing import Any, Optional

from pydantic import Field
from scale_egp.utils.model_utils import BaseModel, Entity, hiddenstr

ModelState = enum.StrEnum(
    "ModelState", [("ENABLED", "ENABLED"), ("PENDING", "PENDING"), ("DISABLED", "DISABLED")]
)
ModelVendor = enum.StrEnum(
    "ModelVendor",
    [
        ("OPENAI", "OPENAI"),
        ("COHERE", "COHERE"),
        ("ANTHROPIC", "ANTHROPIC"),
        ("LLMENGINE", "LLMENGINE"),
        ("OTHER", "OTHER"),
    ],
)
ModelEndpointType = enum.StrEnum(
    "ModelEndpointType",
    [("SYNC", "SYNC"), ("ASYNC", "ASYNC"), ("STREAMING", "STREAMING"), ("BATCH", "BATCH")],
)
ModelType = enum.StrEnum(
    "ModelType",
    [
        ("COMPLETION", "COMPLETION"),
        ("CHAT_COMPLETION", "CHAT_COMPLETION"),
        ("AGENT", "AGENT"),
        ("EMBEDDING", "EMBEDDING"),
        ("RERANKING", "RERANKING"),
    ],
)


class ModelRequest(BaseModel):
    name: str = Field(...)
    model_state: ModelState = Field(...)
    vendor: ModelVendor = Field(...)
    vendor_name: str = Field(...)
    base_model_id: Optional[str] = Field(None)


class Model(Entity):
    """
    Entity for all models, including both self-hosted and 3rd party, base- and fine-tuned models.
    """

    name: str = Field(...)
    model_state: ModelState = Field(...)
    vendor: ModelVendor = Field(...)
    vendor_name: str = Field(...)
    base_model_id: Optional[str] = Field(None)
    id: str = Field(..., description="The unique identifier of the entity.")
    created_at: datetime.datetime = Field(
        ..., description="The date and time when the entity was created in ISO format."
    )
    account_id: str = Field(..., description="The ID of the account that owns the given entity.")
    created_by_user_id: str = Field(..., description="The user who originally created the entity.")


class ScaleModelTemplateRequest(BaseModel):
    name: str = Field(...)
    endpoint_type: ModelEndpointType = Field(...)
    model_type: ModelType = Field(...)
    weights_uri: str = Field(...)
    docker_image: str = Field(...)
    docker_env_vars: Optional[dict[str, Any]] = Field(None)
    launch_parameters: dict[str, Any] = Field(...)


class ScaleModelTemplate(Entity):
    name: str = Field(...)
    endpoint_type: ModelEndpointType = Field(...)
    model_type: ModelType = Field(...)
    weights_uri: str = Field(...)
    docker_image: str = Field(...)
    docker_env_vars: Optional[dict[str, Any]] = Field(None)
    launch_parameters: dict[str, Any] = Field(...)
    id: str = Field(..., description="The unique identifier of the entity.")
    created_at: datetime.datetime = Field(
        ..., description="The date and time when the entity was created in ISO format."
    )
    account_id: str = Field(..., description="The ID of the account that owns the given entity.")
    created_by_user_id: str = Field(..., description="The user who originally created the entity.")


class ScaleModelInstanceRequest(BaseModel):
    model_id: str = Field(...)
    model_template_id: str = Field(...)
    endpoint_url: Optional[hiddenstr] = Field(None)


class ScaleModelInstance(Entity):
    model_id: str = Field(...)
    model_template_id: str = Field(...)
    id: str = Field(..., description="The unique identifier of the entity.")
    created_at: datetime.datetime = Field(
        ..., description="The date and time when the entity was created in ISO format."
    )
    created_by_user_id: str = Field(..., description="The user who originally created the entity.")


class ModelUsageRequest(BaseModel):
    model_id: str = Field(...)
    user_id: str = Field(...)
    started_at: datetime.datetime = Field(...)
    finished_at: Optional[datetime.datetime] = Field(None)


class ModelUsage(Entity):
    model_id: str = Field(...)
    user_id: str = Field(...)
    started_at: datetime.datetime = Field(...)
    finished_at: Optional[datetime.datetime] = Field(None)
    id: str = Field(..., description="The unique identifier of the entity.")
    created_at: datetime.datetime = Field(
        ..., description="The date and time when the entity was created in ISO format."
    )



"""
TODO: module docstring
"""
# Based on packages/egp-api-backend/egp_api_backend/server/internal/exceptions.py
import json
from typing import Optional
import httpx


class EGPException(Exception):
    """
    General EGP SDK exception.
    """

    message: str
    code: int = 500

    def __init__(self, message: str, code: Optional[int] = None):
        self.message = message
        if code is not None:
            self.code = code

    def __str__(self):
        return self.message

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.message}"


class UnsupportedException(EGPException):
    """
    Raised when an unsupported operation is attempted
    """

    code = 400


class ItemDoesNotExist(EGPException):
    """
    Raised when a query expects one row but finds none
    """

    code = 404


def exception_from_response(response: httpx.Response) -> Exception:
    """
    TODO: docstring
    """
    # TODO: raise useful exceptions
    # 404: not found
    # 403: no permission
    # 422: server-side pydantic validation error
    cls = EGPException
    if response.status_code == 404:
        cls = ItemDoesNotExist
    body = response.content.decode("utf-8")
    if body.startswith("{"):  # assume JSON
        try:
            parsed_body = json.loads(body)
            body = parsed_body["detail"]
        except Exception:  # pylint: disable=broad-exception-caught
            pass
    return cls(body, response.status_code)



from __future__ import annotations

from datetime import datetime
from typing import Literal, Optional, Union, List, Dict, Any

from pydantic import Field, validator

from scale_egp.sdk.enums import (
    EmbeddingModelName,
    DataSource,
    DeduplicationStrategy,
    ChunkingStrategy,
    EvaluationType,
    QuestionType,
    ExtraInfoSchema,
    CrossEncoderModelName,
)
from scale_egp.utils.model_utils import BaseModel, Entity

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

class EmbeddingConfig(BaseModel):
    """
    A data model representing the configuration of the embedding.

    Attributes:
        embedding_model: The name of the embedding model to use.
    """

    embedding_model: EmbeddingModelName = Field(
        description="The name of the embedding model to use. We only support the models listed "
        "here so far."
    )


class ArtifactsStatus(BaseModel):
    """
    A data model representing the status of the artifacts in a knowledge base.

    Attributes:
        artifacts_completed: Number of artifacts uploaded successfully.
        artifacts_pending: Number of artifacts awaiting upload.
        artifacts_uploading: Number of artifacts with upload in progress.
        artifacts_failed: Number of artifacts that failed upload.
    """

    artifacts_completed: int
    artifacts_pending: int
    artifacts_uploading: int
    artifacts_failed: int


class ChunksStatus(BaseModel):
    """
    A data model representing the status of the chunks in an artifact.

    Attributes:
        chunks_completed: Number of chunks uploaded successfully
        chunks_pending: Number of chunks awaiting upload
        chunks_failed: Number of chunks that failed upload
    """

    chunks_completed: int
    chunks_pending: int
    chunks_failed: int


class S3DataSourceConfig(BaseModel):
    """
    A data model representing the configuration of a S3 data source.

    Attributes:
        source: The data source type. Must be 's3'.
        s3_bucket: The name of the S3 bucket where the data is stored
        s3_prefix: The prefix of the S3 bucket where the data is stored
        aws_region: The AWS region where the S3 bucket is located
        aws_account_id: The AWS account ID that owns the S3 bucket
    """

    source: Literal[DataSource.S3] = DataSource.S3.value
    s3_bucket: str
    aws_region: str
    aws_account_id: str
    s3_prefix: Optional[str] = None


class SharePointDataSourceConfig(BaseModel):
    """
    A data model representing the configuration of a SharePoint data source.

    Attributes:
        source: The data source type. Must be 'sharepoint'.
        client_id: The client ID associated with this SharePoint site
        tenant_id: The tenant ID that the SharePoint site is within
        site_id: The site ID for this SharePoint site
        folder_path: The nested folder path to read files from the root of the site
        recursive: Whether to recurse through the folder contents
    """

    source: Literal[DataSource.SHAREPOINT] = DataSource.SHAREPOINT.value
    client_id: str
    tenant_id: str
    site_id: str
    folder_path: Optional[str] = ""
    recursive: Optional[bool] = True


class GoogleDriveDataSourceConfig(BaseModel):
    """
    A data model representing the configuration of a Google Drive data source.

    Attributes:
        source: The data source type. Must be 'google_drive'.
        drive_id: The ID of the Google Drive to retrieve contents from
    """

    source: Literal[DataSource.GOOGLE_DRIVE] = DataSource.GOOGLE_DRIVE.value
    drive_id: str


class LocalChunksSourceConfig(BaseModel):
    """
    A data model representing the configuration of a local chunks data source.

    Attributes:
        source: The data source type. Must be 'local_chunks'.
        artifact_name: The file name assigned to the artifact, containing a file extension.
            Adding an extension is mandatory, to allow detecting file types for text extraction.
        artifact_uri: A unique identifier for an artifact within the knowledge base, such as full
            path in a directory or file system.
        deduplication_strategy: Action to take if an artifact with the same name already exists
            in the knowledge base. Can be either Overwrite (default) or Fail.
    """

    source: Literal[DataSource.LOCAL_CHUNKS] = DataSource.LOCAL_CHUNKS.value
    artifact_name: str
    artifact_uri: str
    deduplication_strategy: Optional[DeduplicationStrategy] = DeduplicationStrategy.OVERWRITE


class SharePointDataSourceAuthConfig(BaseModel):
    """
    A data model representing the configuration of a SharePoint data source.

    Attributes:
        source: The data source type. Must be 'sharepoint'.
        client_secret: The secret for the app registration associated with this SharePoint site
    """

    source: Literal[DataSource.SHAREPOINT] = DataSource.SHAREPOINT.value
    client_secret: str


class CharacterChunkingStrategyConfig(BaseModel):
    """
    A data model representing the configuration of a character chunking strategy.

    Attributes:
        strategy: The chunking strategy type. Must be 'character'.
        separator: Character designating breaks in input data. Text data will first be split
            into sections by this separator, then each section will be split into chunks
            of size `chunk_size`.
        chunk_size: Maximum number of characters in each chunk. If not specified, a chunk size
            of 1000 will be used.
        chunk_overlap: Number of characters to overlap between chunks. If not specified, an overlap
            of 200 will be used. For example if the chunk size is 3 and the overlap size
            is 1, and the text to chunk is 'abcde', the chunks will be 'abc', 'cde'.
    """

    strategy: Literal[ChunkingStrategy.CHARACTER] = ChunkingStrategy.CHARACTER.value
    separator: Optional[str] = "\n\n"
    chunk_size: Optional[int] = 1000
    chunk_overlap: Optional[int] = 200


class CategoricalChoice(BaseModel):
    """
    A choice for a categorical question.

    This is only used in a StudioEvaluationConfig to specify a choice for a  question that will be
    asked to users when they are evaluating generated outputs in the EGP annotation UI.

    Attributes:
        label: The text displayed to annotators for this choice.
        value: The value reported in the TestCaseResult for this question if this choice is
            selected.

            If users would like to track the improvement of a model over time, it is
            recommended to use a numeric value for this field.

            A string value can be used if no ordering is desired.
    """

    label: str
    value: Union[str, int, bool]


class CategoricalQuestion(BaseModel):
    """
    A categorical question.

    This is only used in a StudioEvaluationConfig to specify a question that will be asked to
    users when they are evaluating generated outputs in the EGP annotation UI.

    Attributes:
        question_id: A unique id for this question. Each question_id will be represented as a
            key in the `result` field in the TestCaseResult after an annotation is completed.
            When examining how an application has performed on a specific question, users should
            look at the `result` field in the TestCaseResult for the key with the
            corresponding id.
        question_type: The type of the question.
        title: The text displayed to annotators for this question.
        choices: The choices for the question.
    """

    question_id: str
    question_type: Literal[QuestionType.CATEGORICAL] = QuestionType.CATEGORICAL.value
    title: str
    choices: List[CategoricalChoice]


Question = CategoricalQuestion


class StudioEvaluationConfig(BaseModel):
    """
    This specifies the configuration for a studio evaluation job.

    Users should use this evaluation config if they intend to do human evaluation through
    [Studio](https://scale.com/studio).

    Attributes:
        evaluation_type: The type of the evaluation.
        studio_project_id: The ID of the Studio project to use for the evaluation.
        questions: The questions to ask users when they are evaluating generated outputs in the
            EGP annotation UI.
    """

    evaluation_type: EvaluationType = EvaluationType.STUDIO
    studio_project_id: str
    questions: List[Question]


EvaluationConfig = StudioEvaluationConfig


class EvaluationDatasetVersion(Entity):
    num: int
    evaluation_dataset_id: str
    id: str
    created_at: datetime
    account_id: str
    created_by_user_id: str


class ExtraInfo(BaseModel):
    """
    A data model representing the extra info of a generation model.

    Attributes:
        schema_type: The schema type of the extra info
        info: The content of the extra info. This must match the schema type.
    """

    schema_type: ExtraInfoSchema
    info: str

    @validator("info")
    def info_matches_schema_type(cls, info, values):
        schema_type = values.get("schema_type")
        if schema_type == ExtraInfoSchema.STRING:
            if not isinstance(info, str):
                raise ValueError("schema_type STRING requires info to be a string")
        else:
            raise ValueError(f"Unknown schema_type: {schema_type}")
        return info


class GenerationTestCaseData(BaseModel):
    """
    A data model representing the data of a Testcase with a GENERATION schema type.

    Attributes:
        input: The input to the generation model
        expected_output: The expected output of the generation model
        expected_extra_info: The expected extra info of the generation model
    """

    input: str
    expected_output: Optional[str] = ""
    expected_extra_info: Optional[ExtraInfo] = ExtraInfo(
        schema_type=ExtraInfoSchema.STRING,
        info="",
    )


TestCaseData: TypeAlias = GenerationTestCaseData


class GenerationTestCaseResultData(BaseModel):
    """
    A data model representing the data of a TestcaseResult with a GENERATION schema type.

    Attributes:
        generation_output: The output of the generation model
        generation_extra_info: The extra info of the generation model
    """

    generation_output: str
    generation_extra_info: Optional[ExtraInfo] = None


TestCaseResultData: TypeAlias = GenerationTestCaseResultData


class CompletionContent(BaseModel):
    """
    A data model representing the completion text and the finish reason.

    Attributes:
        text: Completion text. If streaming, this field will contain each packet of text.
        finish_reason: Reason the LLM finished generating text.
    """

    text: str
    finish_reason: Optional[str] = None


class TokenUsage(BaseModel):
    """
    A data model representing LLM token usage numbers.

    Attributes:
        prompt: Number of tokens in the prompt.
        completion: Number of tokens in the completion.
        total: Total number of tokens in both the prompt and the completion.
    """

    prompt: Optional[int] = None
    completion: Optional[int] = None
    total: int


class Completion(BaseModel):
    """
    A data model representing a completion.

    Attributes:
        completion: The actual completion text and the finish reason.
        token_usage: Token usage numbers. If streaming, this field is null until the stream
            completes, at which point it will be populated (if supported).
    """

    completion: CompletionContent
    token_usage: Optional[TokenUsage] = None


class ModelParameters(BaseModel):
    """
    A data model representing the configuration parameters for the completion model.

    Attributes:
        temperature: What sampling temperature to use, between [0, 1]. Higher values like 0.8
            will make the output more random, while lower values like 0.2 will make it more
            focused and deterministic. Setting temperature=0.0 will enable fully deterministic
            (greedy) sampling.
        stop_sequences: List of up to 4 sequences where the API will stop generating further
            tokens. The returned text will not contain the stop sequence.
        max_tokens: The maximum number of tokens to generate in the completion. The token count
            of your prompt plus max_tokens cannot exceed the model's context length. If not,
            specified, max_tokens will be determined based on the model used:
            | Model API family | Model API default | EGP applied default |
            | --- | --- | --- |
            | OpenAI Completions | [`16`](
            https://platform.openai.com/docs/api-reference/completions/create#max_tokens) |
            `context window - prompt size` |
            | OpenAI Chat Completions | [`context window - prompt size`](
            https://platform.openai.com/docs/api-reference/chat/create#max_tokens) | `context
            window - prompt size` |
    """

    temperature: float = Field(default=0.2, ge=0.0, le=1.0)
    stop_sequences: Optional[List[str]] = Field(default=None, max_items=4)
    max_tokens: Optional[int] = Field(default=None)


class ChunkToUpload(BaseModel):
    """
    A data model representing a local chunk.

    Attributes:
        text: The text associated with the chunk
        chunk_position: The position of the chunk in the artifact
        metadata: Any additional key value pairs of information stored with the chunk
    """

    text: str
    chunk_position: int
    metadata: Dict[str, Any] = Field(default_factory=dict)


class CrossEncoderRankParams(BaseModel):
    cross_encoder_model: Literal[
        "cross-encoder/ms-marco-MiniLM-L-12-v2",
        "cross-encoder/mmarco-mMiniLMv2-L12-H384-v1",
    ] = Field(
        default=CrossEncoderModelName.CROSS_ENCODER_MS_MARCO_MINILM_L12_V2,
        description="Cross encoder model to use when ranking.",
    )


class CrossEncoderRankStrategy(BaseModel):
    method: Literal["cross_encoder"] = Field(
        default="cross_encoder",
        const=True,
        description="The name of the rank strategy. Must be `cross_encoder`.",
    )
    params: CrossEncoderRankParams = Field(
        default=CrossEncoderRankParams(),
        description="The parameters needed for ranking.",
    )


class RougeRankParams(BaseModel):
    metric: str = Field(
        default="rouge2",
        description="Rouge type, can be n-gram based (e.g. rouge1, rouge2) or longest common "
        "subsequence (rougeL or rougeLsum)",
    )
    score: Literal["precision", "recall", "fmeasure"] = Field(
        default="recall", description="Metric to use from Rouge score"
    )


class RougeRankStrategy(BaseModel):
    method: Literal["rouge"] = Field(
        default="rouge",
        const=True,
        description="The name of the rank strategy.",
    )
    params: RougeRankParams = Field(
        default=RougeRankParams(),
        description="The parameters needed for ranking.",
    )



import functools
import json
from typing import (
    Any,
    List,
    Optional,
    TYPE_CHECKING,
    Dict,
    Iterable,
    Callable,
)
from urllib.parse import urljoin

from httpx import Response

from scale_egp.exceptions import exception_from_response
from scale_egp.utils.model_utils import BaseModel

if TYPE_CHECKING:
    from scale_egp.sdk.client import EGPClient

DEFAULT_TIMEOUT = 60


def handle_api_exceptions(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        response = func(*args, **kwargs)
        if response.status_code != 200:
            raise exception_from_response(response)
        return response

    return wrapper


class APIEngine:
    def __init__(self, api_client: "EGPClient"):
        self._api_client = api_client

    def _post(
        self,
        sub_path: str,
        request: Optional[BaseModel] = None,
        timeout: Optional[int] = None,
    ) -> Response:
        response = self._raw_post(
            sub_path=sub_path,
            request_json=request.model_dump() if request is not None else None,
            timeout=timeout if timeout is not None else DEFAULT_TIMEOUT,
        )
        return response

    def _post_stream(
        self,
        sub_path: str,
        request: Optional[BaseModel] = None,
        timeout: Optional[int] = None,
    ) -> Iterable[Dict[str, Any]]:
        response = self._raw_stream(
            sub_path=sub_path,
            request_json=request.model_dump() if request is not None else None,
            timeout=timeout if timeout is not None else DEFAULT_TIMEOUT,
        )
        with response as lines:
            for line in lines.iter_lines():
                if line.startswith('data: '):
                    event_json_str = line[len('data: '):]
                    try:
                        yield json.loads(event_json_str)
                    except json.JSONDecodeError:
                        raise ValueError(f"Invalid JSON payload: {event_json_str}")

    def _post_batch(
        self,
        sub_path: str,
        request_batch: Optional[List[BaseModel]] = None,
        timeout: Optional[int] = None,
    ) -> Response:
        response = self._raw_post(
            sub_path=sub_path,
            request_json=[request.model_dump() for request in request_batch],
            timeout=timeout if timeout is not None else DEFAULT_TIMEOUT,
        )
        return response

    def _get(
        self,
        sub_path: str,
        query_params: Optional[Dict[str, str]] = None,
        timeout: Optional[int] = None,
    ) -> Response:
        response = self._raw_get(
            sub_path=sub_path,
            query_params=query_params,
            timeout=timeout if timeout is not None else DEFAULT_TIMEOUT,
        )
        return response

    def _patch(
        self,
        sub_path: str,
        request: Optional[BaseModel] = None,
        timeout: Optional[int] = None,
    ) -> Response:
        response = self._raw_patch(
            sub_path=sub_path,
            request_json=request.model_dump() if request is not None else None,
            timeout=timeout if timeout is not None else DEFAULT_TIMEOUT,
        )
        return response

    def _delete(
        self,
        sub_path: str,
        timeout: Optional[int] = None,
    ) -> Response:
        response = self._raw_delete(
            sub_path=sub_path,
            timeout=timeout if timeout is not None else DEFAULT_TIMEOUT,
        )
        return response

    @handle_api_exceptions
    def _raw_post(
        self,
        sub_path: str,
        request_json: Optional[Dict[str, Any]] = None,
        timeout: Optional[int] = None,
        additional_headers: Optional[Dict[str, str]] = None,
    ) -> Response:
        response = self._api_client.httpx_client.post(
            urljoin(self._api_client.endpoint_url, sub_path),
            json=request_json,
            **self._universal_request_kwargs(
                timeout=timeout,
                additional_headers=additional_headers,
            ),
        )
        return response

    def _raw_stream(
        self,
        sub_path: str,
        request_json: Optional[Dict[str, Any]] = None,
        timeout: Optional[int] = None,
        additional_headers: Optional[Dict[str, str]] = None,
        method: str = "POST",
    ):
        return self._api_client.httpx_client.stream(
            method=method,
            url=urljoin(self._api_client.endpoint_url, sub_path),
            json=request_json,
            timeout=timeout if timeout is not None else DEFAULT_TIMEOUT,
            headers={
                **(additional_headers if additional_headers is not None else {}),
            },
        )

    @handle_api_exceptions
    def _raw_get(
        self,
        sub_path: str,
        timeout: Optional[int] = None,
        query_params: Optional[Dict[str, str]] = None,
        additional_headers: Optional[Dict[str, str]] = None,
    ) -> Response:
        response = self._api_client.httpx_client.get(
            urljoin(self._api_client.endpoint_url, sub_path),
            params=query_params,
            **self._universal_request_kwargs(
                timeout=timeout,
                additional_headers=additional_headers,
            ),
        )
        return response

    @handle_api_exceptions
    def _raw_patch(
        self,
        sub_path: str,
        request_json: Optional[Dict[str, Any]] = None,
        timeout: Optional[int] = None,
        additional_headers: Optional[Dict[str, str]] = None,
    ) -> Response:
        response = self._api_client.httpx_client.patch(
            urljoin(self._api_client.endpoint_url, sub_path),
            json=request_json,
            **self._universal_request_kwargs(
                timeout=timeout,
                additional_headers=additional_headers,
            ),
        )
        return response

    @handle_api_exceptions
    def _raw_delete(
        self,
        sub_path: str,
        timeout: Optional[int] = None,
        additional_headers: Optional[Dict[str, str]] = None,
    ) -> Response:
        response = self._api_client.httpx_client.delete(
            url=urljoin(self._api_client.endpoint_url, sub_path),
            **self._universal_request_kwargs(
                timeout=timeout,
                additional_headers=additional_headers,
            ),
        )
        return response

    def _universal_request_kwargs(
        self,
        timeout: Optional[int] = None,
        additional_headers: Optional[Dict[str, str]] = None,
    ) -> Dict[str, Any]:
        return dict(
            timeout=timeout if timeout is not None else DEFAULT_TIMEOUT,
            headers={
                **(additional_headers if additional_headers is not None else {}),
            },
        )



from typing import List, Optional, Dict, Any, Union

import httpx

from scale_egp.sdk.enums import EmbeddingModelName, ChunkUploadStatus
from scale_egp.sdk.models import (
    KnowledgeBase,
    KnowledgeBaseRequest,
    EmbeddingConfig,
    Chunk,
    KnowledgeBaseQueryRequest,
    KnowledgeBaseUpload,
    LocalChunksSourceConfig,
    ChunkToUpload,
    KnowledgeBaseArtifact,
    S3DataSourceConfig,
    GoogleDriveDataSourceConfig,
    SharePointDataSourceConfig,
    SharePointDataSourceAuthConfig,
    CharacterChunkingStrategyConfig,
    KnowledgeBaseResponse,
    ListKnowledgeBasesResponse,
    ListKnowledgeBaseArtifactsResponse,
    KnowledgeBaseQueryResponse,
    KnowledgeBaseRemoteUploadRequest,
    KnowledgeBaseUploadResponse,
    KnowledgeBaseLocalChunkUploadRequest,
    ListKnowledgeBaseUploadsResponse,
    CancelKnowledgeBaseUploadResponse,
)
from scale_egp.utils.api_utils import APIEngine


class KnowledgeBaseCollection(APIEngine):

    _sub_path = "v2/knowledge-bases"

    def uploads(self) -> "KnowledgeBaseUploadsCollection":
        """
        Returns a KnowledgeBaseUploadsCollection object for uploads associated with a knowledge
        base.

        Returns:
            A KnowledgeBaseUploadsCollection object.
        """
        return KnowledgeBaseUploadsCollection(self._api_client)

    def artifacts(self) -> "KnowledgeBaseArtifactsCollection":
        """
        Returns a KnowledgeBaseArtifactsCollection object for artifacts associated with a
        knowledge base.

        Returns:
            A KnowledgeBaseArtifactsCollection object.
        """
        return KnowledgeBaseArtifactsCollection(self._api_client)

    def create(
        self,
        name: str,
        embedding_model_name: EmbeddingModelName,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> KnowledgeBase:
        """
        Create a new Knowledge Base.

        Args:
            name: The name of the Knowledge Base.
            embedding_model_name: The name of the embedding model to use for the Knowledge Base.
            metadata: The metadata of the Knowledge Base.

        Returns:
            The newly created Knowledge Base.
        """
        response = self._post(
            sub_path=self._sub_path,
            request=KnowledgeBaseRequest(
                account_id=self._api_client.account_id,
                knowledge_base_name=name,
                embedding_config=EmbeddingConfig(
                    embedding_model=embedding_model_name,
                ),
                metadata=metadata,
            ),
        )
        response_model = KnowledgeBaseResponse.from_dict(response.json())
        return self.get(id=response_model.knowledge_base_id)

    def get(
        self,
        id: str,
    ) -> KnowledgeBase:
        """
        Get an Knowledge Base by ID.

        Args:
            id: The ID of the Knowledge Base.

        Returns:
            The Knowledge Base.
        """
        response = self._get(
            sub_path=f"{self._sub_path}/{id}",
        )
        return KnowledgeBase.from_dict(response.json())

    def delete(
        self,
        id: str,
    ) -> bool:
        """
        Delete a Knowledge Base by ID.

        Args:
            id: The ID of the Knowledge Base.
        """
        response = self._delete(
            sub_path=f"{self._sub_path}/{id}",
        )
        return response.status_code == httpx.codes.ok

    def list(
        self,
    ) -> List[KnowledgeBase]:
        """
        List all Knowledge Bases.

        Returns:
            A list of Knowledge Bases.
        """
        response = self._get(
            sub_path=self._sub_path,
        )
        response_model = ListKnowledgeBasesResponse.from_dict(response.json())
        return response_model.items

    def query(
        self,
        knowledge_base: KnowledgeBase,
        query: str,
        top_k: int,
        include_embeddings: bool = False,
    ) -> List[Chunk]:
        """
        Query a Knowledge Base.

        Args:
            knowledge_base: The Knowledge Base to query.
            query: The query string.
            top_k: The number of results to return.
            include_embeddings: Whether to include embeddings in the response.

        Returns:
            The query response.
        """
        response = self._post(
            sub_path=f"{self._sub_path}/{knowledge_base.knowledge_base_id}/query",
            request=KnowledgeBaseQueryRequest(
                query=query,
                top_k=top_k,
                include_embeddings=include_embeddings,
            ),
        )
        response_model = KnowledgeBaseQueryResponse.from_dict(response.json())
        return response_model.chunks


class KnowledgeBaseArtifactsCollection(APIEngine):

    _sub_path = "v2/knowledge-bases/{knowledge_base_id}/artifacts"

    def get(
        self,
        id: str,
        knowledge_base: KnowledgeBase,
        status_filter: Optional[ChunkUploadStatus] = ChunkUploadStatus.COMPLETED.value,
    ) -> KnowledgeBaseArtifact:
        """
        Get a Knowledge Base Artifact by ID.

        Args:
            id: The ID of the Knowledge Base Artifact.
            knowledge_base: The Knowledge Base the artifact was created for.
            status_filter: Return only artifacts with the given status.

        Returns:
            The Knowledge Base Artifact.
        """
        response = self._get(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}/"
            f"{id}",
            query_params=dict(
                status_filter=status_filter,
            ),
        )
        return KnowledgeBaseArtifact.from_dict(response.json())

    def list(
        self,
        knowledge_base: KnowledgeBase,
    ) -> List[KnowledgeBaseArtifact]:
        """
        List all Knowledge Base Artifacts.

        Returns:
            A list of Knowledge Base Artifacts.
        """
        response = self._get(
            sub_path=self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id),
        )
        response_model = ListKnowledgeBaseArtifactsResponse.from_dict(response.json())
        return response_model.artifacts


class KnowledgeBaseUploadsCollection(APIEngine):
    _sub_path = "v2/knowledge-bases/{knowledge_base_id}/uploads"

    def create_remote_upload(
        self,
        knowledge_base: KnowledgeBase,
        data_source_config: Union[
            S3DataSourceConfig, SharePointDataSourceConfig, GoogleDriveDataSourceConfig
        ],
        data_source_auth_config: Optional[Union[SharePointDataSourceAuthConfig]],
        chunking_strategy_config: Union[CharacterChunkingStrategyConfig],
    ) -> KnowledgeBaseUpload:
        """
        Create a new remote upload.

        Args:
            knowledge_base: The Knowledge Base to upload data to.
            data_source_config: The data source config.
            data_source_auth_config: The data source auth config.
            chunking_strategy_config: The chunking strategy config.

        Returns:
            The newly created remote upload.
        """
        response = self._post(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}",
            request=KnowledgeBaseRemoteUploadRequest(
                data_source_config=data_source_config,
                data_source_auth_config=data_source_auth_config,
                chunking_strategy_config=chunking_strategy_config,
            ),
        )
        response_model = KnowledgeBaseUploadResponse.from_dict(response.json())
        return self.get(id=response_model.upload_id, knowledge_base=knowledge_base)

    def create_local_upload(
        self,
        knowledge_base: KnowledgeBase,
        data_source_config: LocalChunksSourceConfig,
        chunks: List[ChunkToUpload],
    ) -> KnowledgeBaseUpload:
        """
        Create a new local upload.

        Args:
            knowledge_base: The Knowledge Base to upload data to.
            data_source_config: The data source config.
            chunks: The chunks to upload.

        Returns:
            The newly created local upload.
        """
        response = self._post(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}",
            request=KnowledgeBaseLocalChunkUploadRequest(
                data_source_config=data_source_config,
                chunks=chunks,
            ),
        )
        response_model = KnowledgeBaseUploadResponse.from_dict(response.json())
        return self.get(id=response_model.upload_id, knowledge_base=knowledge_base)

    def get(
        self,
        id: str,
        knowledge_base: KnowledgeBase,
    ) -> KnowledgeBaseUpload:
        """
        Get an Knowledge Base Upload by ID.

        Args:
            id: The ID of the Knowledge Base Upload.
            knowledge_base: The Knowledge Base the upload was created for.

        Returns:
            The Knowledge Base Upload.
        """
        response = self._get(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}/"
            f"{id}",
        )
        return KnowledgeBaseUpload.from_dict(response.json())

    def list(
        self,
        knowledge_base: KnowledgeBase,
    ) -> List[KnowledgeBaseUpload]:
        """
        List all Knowledge Base Uploads.

        Returns:
            A list of Knowledge Base Uploads.
        """
        response = self._get(
            sub_path=self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id),
        )
        response_model = ListKnowledgeBaseUploadsResponse.from_dict(response.json())
        return response_model.uploads

    def cancel(
        self,
        knowledge_base: KnowledgeBase,
        id: str,
    ) -> bool:
        """
        Cancel an upload.

        Args:
            knowledge_base: The Knowledge Base the upload was created for.
            id: The ID of the upload to cancel.

        Returns:
            True if the upload was canceled, False otherwise.
        """
        response = self._post(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}/"
            f"{id}/cancel",
            request=None,
        )
        response_model = CancelKnowledgeBaseUploadResponse.from_dict(response.json())
        return response_model.canceled



import os
from typing import Dict, Optional, Union, Any

import httpx
from httpx import Client

from scale_egp.sdk.collections.application_specs import ApplicationSpecCollection
from scale_egp.sdk.collections.chunks import ChunkCollection
from scale_egp.sdk.collections.completions import CompletionCollection
from scale_egp.sdk.collections.evaluation_datasets import EvaluationDatasetCollection
from scale_egp.sdk.collections.evaluations import EvaluationCollection
from scale_egp.sdk.collections.knowledge_bases import KnowledgeBaseCollection
from scale_egp.sdk.collections.studio_projects import StudioProjectCollection

DEFAULT_ENDPOINT_URL = "https://api.egp.scale.com"


class EGPClient:
    """
    The EGP client object. This is the main entry point for interacting with EGP.

    From this client you can access "collections" which interact with various EGP components.
    Each collections will have various methods that interact with the API. Some collections may
    have sub-collections to signify a hierarchical relationship between the entities they represent.
    """

    def __init__(
        self,
        api_key: str = None,
        account_id: str = None,
        endpoint_url: str = None,
        proxies: Dict[str, str] = None,
        auth: Optional[httpx.Auth] = None,
        verify: Optional[Union[str, bool]] = False,
        **kwargs: Dict[str, Any],
    ):
        """
        Args:
            api_key: The EGP API key to use. If not provided, the `EGP_API_KEY` environment
                variable will be used. Enterprise customers of EGP should use the API key provided
                to them by their Scale account manager.
            account_id: The EGP account ID to use. If not provided, the `ACCOUNT_ID` environment
                variable will be used.
            endpoint_url: The EGP endpoint URL to use. If not provided, the `EGP_ENDPOINT_URL`
                environment variable will be used. If that is not set, the default EGP endpoint
                URL `https://api.egp.scale.com` will be used. Enterprise customers of EGP should
                use the endpoint URL provided by their Scale account manager.
            proxies: The proxies to use for making requests. Dictionary mapping protocol to the
                URL of the proxy. See
                [HTTPX docs](https://www.python-httpx.org/advanced/#http-proxying) for more
                information.
            auth: The authentication mechanism to use for making requests. We use HTTPX for
                requests, so please reference the
                [HTTPX docs](https://www.python-httpx.org/quickstart/#authentication)
                for more information.
            verify: Either a boolean, in which case it controls whether we verify the server's TLS
                certificate, or a string, in which case it must be a path to a CA bundle to use.
                See: https://www.python-httpx.org/advanced/#ssl-certificates
            **kwargs: Further arguments to pass to the httpx client
        """
        api_key = api_key or os.environ.get("EGP_API_KEY")
        account_id = account_id or os.environ.get("ACCOUNT_ID")
        endpoint_url = endpoint_url or os.environ.get("EGP_ENDPOINT_URL", DEFAULT_ENDPOINT_URL)

        self.api_key = api_key
        self.account_id = account_id
        self.endpoint_url = endpoint_url if endpoint_url.endswith('/') else endpoint_url + '/'
        self.proxies = proxies
        self.auth = auth
        self.verify = verify

        if not self.api_key:
            raise ValueError("No API key provided. Please provide an API key.")
        if not self.endpoint_url:
            raise ValueError("No endpoint URL provided. Please provide an endpoint URL.")

        self.httpx_client = Client(
            headers={"x-api-key": self.api_key},
            proxies=self.proxies,
            auth=self.auth,
            verify=self.verify,
            **kwargs,
        )

    def knowledge_bases(self) -> KnowledgeBaseCollection:
        """
        Returns the Knowledge Base Collection.

        Use this collection to create and manage Knowledge Bases.

        Returns:
            The Knowledge Base Collection.
        """
        return KnowledgeBaseCollection(self)

    def chunks(self) -> ChunkCollection:
        """
        Returns the Chunk Collection.

        Use this collection to create and manage Chunks.

        Returns:
            The Chunk Collection.
        """
        return ChunkCollection(self)

    def completions(self) -> CompletionCollection:
        """
        Returns the Completion Collection.

        Use this collection if you want to make request to an LLM to generate a completion.
        """
        return CompletionCollection(self)

    def evaluation_datasets(self) -> EvaluationDatasetCollection:
        """
        Returns the Evaluation Dataset Collection.

        Use this collection to create and manage Evaluation Datasets or Test Cases within them.

        Returns:
            The Evaluation Dataset Collection.
        """
        return EvaluationDatasetCollection(self)

    def studio_projects(self) -> StudioProjectCollection:
        """
        Returns the Studio Project Collection.

        Use this collection to create and manage Studio Projects. These are projects that will be
        used to annotate data in the Scale [Studio](https://scale.com/studio) platform.

        Returns:
            The Studio Project Collection.
        """
        return StudioProjectCollection(self)

    def application_specs(self) -> ApplicationSpecCollection:
        """
        Returns the Application Spec Collection.

        Use this collection to create and manage Application Specs. These are specifications for
        the AI application you are building. They contain information about the AI application
        such as its name and description. They are useful to associate your Evaluations with so
        evaluations can be grouped by application.

        Returns:
            The Application Spec Collection.
        """
        return ApplicationSpecCollection(self)

    def evaluations(self) -> EvaluationCollection:
        """
        Returns the Evaluation Collection.

        Use this collection to create and manage Evaluations and Test Case Results.

        Evaluations are used to evaluate the performance of AI applications. Users are
        expected to follow the following procedure to perform an evaluation:

        1. Select an Evaluation Dataset
        2. Iterate through the dataset's Test Cases:
          - For each of these test cases, the user use their AI application to generate output
          data on each test case input prompt.
        3. The user then submits this data as as batch of Test Case Results associated
        with an Evaluation.
        4. Annotators will asynchronously log into the Scale [Studio](https://scale.com/studio)
        platform to annotate the submitted Test Case Results. The annotations will be used to
        evaluate the performance of the AI application.
        5. The submitting user will check back on their Test Case Results to see if the `result`
        field was populated. If so, the evaluation is complete and the user can use the annotation
        data to evaluate the performance of their AI application.

        Returns:
            The Evaluation Collection.
        """
        return EvaluationCollection(self)
